import sqlalchemy
from sqlalchemy import inspect, or_, cast

from jet_bridge_base.filters.char_filter import CharFilter
from jet_bridge_base.filters.filter import EMPTY_VALUES


def search_queryset(qs, mapper, search):
    if search in EMPTY_VALUES:
        return qs

    def map_column(column):
        if isinstance(column.type, (sqlalchemy.Integer, sqlalchemy.Numeric)):
            return cast(column, sqlalchemy.String).__eq__(search)
        elif isinstance(column.type, (sqlalchemy.JSON, sqlalchemy.Enum)):
            return cast(column, sqlalchemy.String).ilike('%{}%'.format(search))
        elif isinstance(column.type, sqlalchemy.String):
            return column.ilike('%{}%'.format(search))

    operators = list(filter(lambda x: x is not None, map(map_column, mapper.columns)))

    return qs.filter(or_(*operators))


def get_model_search_filter(Model):
    mapper = inspect(Model)

    class ModelSearchFilter(CharFilter):
        def filter(self, qs, value):
            value = self.clean_value(value)
            return search_queryset(qs, mapper, value)

    return ModelSearchFilter



from __future__ import absolute_import
import json

from jet_bridge_base import encoders
from jet_bridge_base.responses.base import Response


class JSONResponse(Response):
    headers = {'Content-Type': 'application/json'}
    encoder_class = encoders.JSONEncoder

    def render(self):
        if self.data is None:
            return

        return json.dumps(
            self.data,
            cls=self.encoder_class
        )



from fnmatch import fnmatch
import requests

from jet_bridge_base import settings
from jet_bridge_base.db import get_conf
from jet_bridge_base.utils.async_exec import as_future


def track_database(request):
    if not settings.TRACK_DATABASES_ENDPOINT:
        return

    track_databases = list(filter(lambda x: x != '', map(lambda x: x.lower().strip(), settings.TRACK_DATABASES.split(','))))
    conf = get_conf(request)
    hostname = '{}:{}'.format(conf.get('host', ''), conf.get('port', '')).lower()

    if any(map(lambda x: fnmatch(hostname, x), track_databases)):
        headers = {}
        data = {
            'databaseName': conf.get('name', ''),
            'databaseSchema': conf.get('schema', ''),
            'databaseHost': conf.get('host', ''),
            'databasePort': conf.get('port', '')
        }

        if settings.TRACK_DATABASES_AUTH:
            headers['Authorization'] = settings.TRACK_DATABASES_AUTH

        requests.request('POST', settings.TRACK_DATABASES_ENDPOINT, headers=headers, json=data)


def track_database_async(request):
    if not settings.TRACK_DATABASES_ENDPOINT:
        return

    try:
        import asyncio
    except:
        return

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    as_future(lambda: track_database(request))



from sqlalchemy.sql.base import _bind_or_error
from sqlalchemy import Table
from sqlalchemy import exc
from sqlalchemy import inspection
from sqlalchemy import util

from jet_bridge_base.logger import logger


def reflect(
    metadata,
    bind=None,
    schema=None,
    views=False,
    only=None,
    extend_existing=False,
    autoload_replace=True,
    resolve_fks=True,
    pending_connection=None,
    **dialect_kwargs
):
    if bind is None:
        bind = _bind_or_error(metadata)

    with inspection.inspect(bind)._inspection_context() as insp:
        reflect_opts = {
            "autoload_with": insp,
            "extend_existing": extend_existing,
            "autoload_replace": autoload_replace,
            "resolve_fks": resolve_fks,
            "_extend_on": set(),
        }

        reflect_opts.update(dialect_kwargs)

        if schema is None:
            schema = metadata.schema

        if schema is not None:
            reflect_opts["schema"] = schema

        available = util.OrderedSet(insp.get_table_names(schema))
        if views:
            available.update(insp.get_view_names(schema))

        if schema is not None:
            available_w_schema = util.OrderedSet(
                ["%s.%s" % (schema, name) for name in available]
            )
        else:
            available_w_schema = available

        current = set(metadata.tables)

        if only is None:
            load = [
                name
                for name, schname in zip(available, available_w_schema)
                if extend_existing or schname not in current
            ]
        elif callable(only):
            load = [
                name
                for name, schname in zip(available, available_w_schema)
                if (extend_existing or schname not in current)
                   and only(name, metadata)
            ]
        else:
            missing = [name for name in only if name not in available]
            if missing:
                s = schema and (" schema '%s'" % schema) or ""
                raise exc.InvalidRequestError(
                    "Could not reflect: requested table(s) not available "
                    "in %r%s: (%s)" % (bind.engine, s, ", ".join(missing))
                )
            load = [
                name
                for name in only
                if extend_existing or name not in current
            ]

        """
        Modified: Added default PK set and progress display
        """

        if pending_connection:
            pending_connection['tables_total'] = len(load)

        i = 0
        for name in load:
            try:
                logger.info('Analyzing table "{}" ({} / {})"...'.format(name, i + 1, len(load)))
                table = Table(name, metadata, **reflect_opts)

                args = []
                has_pk = False
                first_column = None

                for item in table.columns:
                    if not has_pk and item.primary_key:
                        has_pk = True
                    if first_column is None:
                        first_column = item

                if not has_pk and first_column is not None:
                    logger.warning('Table "{}" is missing PK: "{}" column was set as PK'.format(name, first_column.name))

                    first_column.primary_key = True
                    args.append(first_column)
                    reflect_opts['extend_existing'] = True
                    table = Table(name, metadata, *args, **reflect_opts)
                    setattr(table, '__jet_auto_pk__', True)


            except exc.UnreflectableTableError as uerr:
                util.warn("Skipping table %s: %s" % (name, uerr))

            i += 1

            if pending_connection:
                pending_connection['tables_processed'] = i

    """
    Modify END
    """



import json

import requests

from jet_bridge_base import settings
from jet_bridge_base.encoders import JSONEncoder
from jet_bridge_base.utils.async_exec import as_future


def track_model(request, model, action, uid, model_data):
    if not settings.TRACK_MODELS_ENDPOINT:
        return

    if request.project is None or request.environment is None or request.resource_token is None:
        return

    url = '{}/model_change'.format(settings.TRACK_MODELS_ENDPOINT)
    headers = {
        'Content-Type': 'application/json'
    }
    data = {
        'project': request.project,
        'environment': request.environment,
        'resource_token': request.resource_token,
        'model': model,
        'action': action,
        'data': model_data
    }

    if uid is not None:
        data['id'] = uid

    if settings.TRACK_MODELS_AUTH:
        headers['Authorization'] = settings.TRACK_MODELS_AUTH

    data_str = json.dumps(data, cls=JSONEncoder)
    requests.post(url, data=data_str, headers=headers)


def track_model_async(request, model, action, uid, data):
    if not settings.TRACK_MODELS_ENDPOINT:
        return

    try:
        import asyncio
    except:
        return

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    as_future(lambda: track_model(request, model, action, uid, data))



import six
from sqlalchemy import inspect
from sqlalchemy.exc import SQLAlchemyError

from jet_bridge_base import fields
from jet_bridge_base.serializers.serializer import Serializer
from jet_bridge_base.utils.exceptions import validation_error_from_database_error

data_types = [
    {'query': 'VARCHAR', 'operator': 'startswith', 'data_type': fields.CharField},
    {'query': 'TEXT', 'operator': 'equals', 'data_type': fields.CharField},
    {'query': 'BOOLEAN', 'operator': 'equals', 'data_type': fields.BooleanField},
    {'query': 'INTEGER[]', 'operator': 'startswith', 'data_type': fields.ArrayField},
    {'query': 'INTEGER', 'operator': 'equals', 'data_type': fields.IntegerField},
    {'query': 'SMALLINT', 'operator': 'equals', 'data_type': fields.IntegerField},
    {'query': 'NUMERIC', 'operator': 'startswith', 'data_type': fields.CharField},
    {'query': 'VARCHAR', 'operator': 'startswith', 'data_type': fields.CharField},
    {'query': 'TIMESTAMP', 'operator': 'startswith', 'data_type': fields.DateTimeField},
    {'query': 'DATETIME', 'operator': 'startswith', 'data_type': fields.DateTimeField},
    {'query': 'JSON', 'operator': 'startswith', 'data_type': fields.JSONField},
    {'query': 'ARRAY', 'operator': 'equals', 'data_type': fields.JSONField},
    {'query': 'geometry', 'operator': 'startswith', 'data_type': fields.WKTField},
    {'query': 'geography', 'operator': 'startswith', 'data_type': fields.WKTField},
]
default_data_type = fields.CharField


def get_column_data_type(column):
    try:
        data_type = six.text_type(column.type)
    except:
        data_type = 'NullType'

    for rule in data_types:
        if rule['operator'] == 'equals' and data_type == rule['query']:
            return rule['data_type']
        elif rule['operator'] == 'startswith' and data_type[:len(rule['query'])] == rule['query']:
            return rule['data_type']

    return default_data_type


class ModelSerializer(Serializer):

    def __init__(self, *args, **kwargs):
        super(ModelSerializer, self).__init__(*args, **kwargs)
        self.session = kwargs.get('context', {}).get('session', None)
        self.model = self.meta.model

    def get_fields(self):
        result = super(ModelSerializer, self).get_fields()

        if hasattr(self.meta, 'model_fields'):
            mapper = inspect(self.meta.model)
            columns = dict(map(lambda x: (x.key, x), mapper.columns))

            for field_name in self.meta.model_fields:
                column = columns.get(field_name)
                data_type = get_column_data_type(column)
                kwargs = {}

                if column.primary_key and column.autoincrement:
                    kwargs['read_only'] = True
                if column.autoincrement or column.default or column.server_default or column.nullable:
                    kwargs['required'] = False

                field = data_type(**kwargs)
                field.field_name = field_name
                result.append(field)

        return result

    def create_instance(self, validated_data):
        ModelClass = self.meta.model
        return ModelClass(**validated_data)

    def create(self, validated_data):
        instance = self.create_instance(validated_data)
        self.session.add(instance)

        try:
            self.session.commit()
        except SQLAlchemyError as e:
            self.session.rollback()
            raise validation_error_from_database_error(e, self.model)

        return instance

    def update(self, instance, validated_data):
        for attr, value in validated_data.items():
            setattr(instance, attr, value)

        try:
            self.session.commit()
        except SQLAlchemyError as e:
            self.session.rollback()
            raise validation_error_from_database_error(e, self.model)

        return instance
